#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from paasta_tools.utils import _run
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        'security-check',
        description='Performs a security check consisting of a few tests.',
        help='Performs a security check',
    )
    list_parser.add_argument(
        '-s', '--service',
        help='Name of service for which you wish to check. Leading "services-", as included in a '
             'Jenkins job name, will be stripped.',
        required=True,
    )
    list_parser.add_argument(
        '-c', '--commit',
        help='Git sha of the image to check',
        required=True,
    )
    list_parser.set_defaults(command=perform_security_check)


def perform_security_check(args):
    """It runs a few security tests, checks the return code and prints output that should help in fixing failures.
    If you are at Yelp, please visit https://confluence.yelpcorp.com/display/PAASTA/PaaSTA+security-check+explained
    to learn more.
    :param args: service - the name of the service; commit - upstream git commit.
    :return: 0 if the security-check passed, non-zero if it failed.
    """
    security_check_command = load_system_paasta_config().get_security_check_command()
    if not security_check_command:
        paasta_print("Nothing to be executed during the security-check step")
        return 0

    command = f"{security_check_command} {args.service} {args.commit}"

    ret_code, output = _run(command, timeout=300, stream=True)
    if ret_code != 0:
        paasta_print(
            "The security-check failed. Please visit y/security-check-runbook to learn how to fix it ("
            "including whitelisting safe versions of packages and docker images).",
        )

    return ret_code



# Copyright 2015-2018 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import difflib
import glob
import os
import re
import subprocess
from string import Formatter
from typing import List
from typing import Tuple

import service_configuration_lib
import yaml
try:
    from yaml.cyaml import CSafeDumper as Dumper
except ImportError:  # pragma: no cover (no libyaml-dev / pypy)
    Dumper = yaml.SafeDumper

from paasta_tools.tron.client import TronClient
from paasta_tools.tron import tron_command_context
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InstanceConfig
from paasta_tools.utils import InvalidInstanceConfig
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import NoDeploymentsAvailable
from paasta_tools.utils import paasta_print

from paasta_tools.monitoring_tools import list_teams
from typing import Optional
from typing import Dict
from typing import Any

MASTER_NAMESPACE = 'MASTER'
SPACER = '.'


class TronNotConfigured(Exception):
    pass


class InvalidTronConfig(Exception):
    pass


class TronConfig(dict):
    """System-level configuration for Tron."""

    def __init__(self, config):
        super().__init__(config)

    def get_cluster_name(self):
        """:returns The name of the Tron cluster"""
        try:
            return self['cluster_name']
        except KeyError:
            raise TronNotConfigured('Could not find name of Tron cluster in system Tron config')

    def get_url(self):
        """:returns The URL for the Tron master's API"""
        try:
            return self['url']
        except KeyError:
            raise TronNotConfigured('Could not find URL of Tron master in system Tron config')


def get_tronfig_folder(cluster, soa_dir):
    return os.path.join(soa_dir, 'tron', cluster)


def load_tron_config():
    return TronConfig(load_system_paasta_config().get_tron_config())


def get_tron_client():
    return TronClient(load_tron_config().get_url())


def compose_instance(job, action):
    return f'{job}{SPACER}{action}'


def decompose_instance(instance):
    """Get (job_name, action_name) from an instance."""
    decomposed = instance.split(SPACER)
    if len(decomposed) != 2:
        raise InvalidInstanceConfig('Invalid instance name: %s' % instance)
    return (decomposed[0], decomposed[1])


class StringFormatter(Formatter):
    def __init__(self, context=None):
        Formatter.__init__(self)
        self.context = context

    def get_value(self, key, args, kwds):
        if isinstance(key, str):
            try:
                return kwds[key]
            except KeyError:
                return self.context[key]
            else:
                return Formatter.get_value(key, args, kwds)


def parse_time_variables(
    command: str,
    parse_time: datetime.datetime = None,
) -> str:
    """Parses an input string and uses the Tron-style dateparsing
    to replace time variables. Currently supports only the date/time
    variables listed in the tron documentation:
    http://tron.readthedocs.io/en/latest/command_context.html#built-in-cc

    :param input_string: input string to be parsed
    :param parse_time: Reference Datetime object to parse the date and time strings, defaults to now.
    :returns: A string with the date and time variables replaced
    """
    if parse_time is None:
        parse_time = datetime.datetime.now()
    # We build up a tron context object that has the right
    # methods to parse tron-style time syntax
    job_context = tron_command_context.JobRunContext(tron_command_context.CommandContext())
    # The tron context object needs the run_time attribute set so it knows
    # how to interpret the date strings
    job_context.job_run.run_time = parse_time
    return StringFormatter(job_context).format(command)


class TronActionConfig(InstanceConfig):
    config_filename_prefix = 'tron'

    def __init__(self, service, instance, cluster, config_dict, branch_dict, soa_dir=DEFAULT_SOA_DIR):
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )
        self.job, self.action = decompose_instance(instance)

    def get_job_name(self):
        return self.job

    def get_action_name(self):
        return self.action

    def get_deploy_group(self) -> Optional[str]:
        return self.config_dict.get('deploy_group', None)

    def get_cmd(self):
        return self.config_dict.get('command')

    def get_executor(self):
        executor = self.config_dict.get('executor', None)
        return 'mesos' if executor == 'paasta' else executor

    def get_healthcheck_mode(self, _) -> None:
        return None

    def get_node(self):
        return self.config_dict.get('node')

    def get_retries(self):
        return self.config_dict.get('retries')

    def get_retries_delay(self):
        return self.config_dict.get('retries_delay')

    def get_requires(self):
        return self.config_dict.get('requires')

    def get_expected_runtime(self):
        return self.config_dict.get('expected_runtime')

    def get_triggered_by(self):
        return self.config_dict.get('triggered_by', None)

    def get_trigger_downstreams(self):
        return self.config_dict.get('trigger_downstreams', None)

    def get_on_upstream_rerun(self):
        return self.config_dict.get('on_upstream_rerun', None)

    def get_trigger_timeout(self):
        return self.config_dict.get('trigger_timeout', None)

    def get_calculated_constraints(self):
        """Combine all configured Mesos constraints."""
        constraints = self.get_constraints()
        if constraints is not None:
            return constraints
        else:
            constraints = self.get_extra_constraints()
            constraints.extend(
                self.get_deploy_constraints(
                    blacklist=self.get_deploy_blacklist(),
                    whitelist=self.get_deploy_whitelist(),
                    # Don't have configs for the paasta cluster
                    system_deploy_blacklist=[],
                    system_deploy_whitelist=None,
                ),
            )
            constraints.extend(self.get_pool_constraints())
            return constraints

    def get_nerve_namespace(self) -> None:
        return None

    def validate(self) -> List[str]:
        # Use InstanceConfig to validate shared config keys like cpus and mem
        error_msgs = super().validate()

        if error_msgs:
            name = self.get_instance()
            return [f'{name}: {msg}' for msg in error_msgs]
        else:
            return []


class TronJobConfig:
    """Represents a job in Tron, consisting of action(s) and job-level configuration values."""

    def __init__(
        self, name: str, config_dict: Dict[str, Any], cluster: str, service: Optional[str] = None,
        load_deployments: bool = True, soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:
        self.name = name
        self.config_dict = config_dict
        self.cluster = cluster
        self.service = service
        self.load_deployments = load_deployments
        self.soa_dir = soa_dir

    def get_name(self):
        return self.name

    def get_node(self):
        return self.config_dict.get('node')

    def get_schedule(self):
        return self.config_dict.get('schedule')

    def get_monitoring(self):
        return self.config_dict.get('monitoring')

    def get_queueing(self):
        return self.config_dict.get('queueing')

    def get_run_limit(self):
        return self.config_dict.get('run_limit')

    def get_all_nodes(self):
        return self.config_dict.get('all_nodes')

    def get_enabled(self):
        return self.config_dict.get('enabled')

    def get_allow_overlap(self):
        return self.config_dict.get('allow_overlap')

    def get_max_runtime(self):
        return self.config_dict.get('max_runtime')

    def get_time_zone(self):
        return self.config_dict.get('time_zone')

    def get_service(self) -> Optional[str]:
        return self.service or self.config_dict.get('service')

    def get_deploy_group(self) -> Optional[str]:
        return self.config_dict.get('deploy_group', None)

    def get_cluster(self):
        return self.cluster

    def get_expected_runtime(self):
        return self.config_dict.get('expected_runtime')

    def _get_action_config(self, action_name, action_dict):
        action_service = action_dict.setdefault('service', self.get_service())
        action_deploy_group = action_dict.setdefault('deploy_group', self.get_deploy_group())
        if action_service and action_deploy_group and self.load_deployments:
            try:
                deployments_json = load_v2_deployments_json(service=action_service, soa_dir=self.soa_dir)
                branch_dict = {
                    'docker_image': deployments_json.get_docker_image_for_deploy_group(action_deploy_group),
                    'git_sha': deployments_json.get_git_sha_for_deploy_group(action_deploy_group),
                    # TODO: add Tron instances when generating deployments json
                    'desired_state': 'start',
                    'force_bounce': None,
                }
            except NoDeploymentsAvailable:
                raise InvalidTronConfig(
                    'No deployment found for action {action} in job {job}, looking for {deploy_group} '
                    'in service {service}'.format(
                        action=action_dict.get('name'),
                        job=self.get_name(),
                        deploy_group=action_deploy_group,
                        service=action_service,
                    ),
                )
        else:
            branch_dict = None

        return TronActionConfig(
            service=action_service,
            instance=compose_instance(self.get_name(), action_name),
            cluster=self.get_cluster(),
            config_dict=action_dict,
            branch_dict=branch_dict,
            soa_dir=self.soa_dir,
        )

    def get_actions(self):
        actions = self.config_dict.get('actions')
        if isinstance(actions, list):
            actions = [
                self._get_action_config(
                    action_dict.get('name'),
                    action_dict,
                )
                for action_dict in actions
            ]
        else:
            actions = [
                self._get_action_config(name, action_dict)
                for name, action_dict in actions.items()
            ]

        return actions

    def get_cleanup_action(self):
        action_dict = self.config_dict.get('cleanup_action')
        if not action_dict:
            return None

        # TODO: we should keep this trickery outside paasta repo
        return self._get_action_config('cleanup', action_dict)

    def check_monitoring(self) -> Tuple[bool, str]:
        monitoring = self.get_monitoring()
        valid_teams = list_teams()
        if monitoring is not None:
            team_name = monitoring.get('team', None)
            if team_name is None:
                return False, 'Team name is required for monitoring'
            elif team_name not in valid_teams:
                suggest_teams = difflib.get_close_matches(word=team_name, possibilities=valid_teams)
                return False, f'Invalid team name: {team_name}. Do you mean one of these: {suggest_teams}'
        return True, ''

    def check_actions(self) -> Tuple[bool, List[str]]:
        actions = self.get_actions()
        cleanup_action = self.get_cleanup_action()
        if cleanup_action:
            actions.append(cleanup_action)

        checks_passed = True
        msgs: List[str] = []
        for action in actions:
            action_msgs = action.validate()
            if action_msgs:
                checks_passed = False
                msgs.extend(action_msgs)
        return checks_passed, msgs

    def validate(self) -> List[str]:
        _, error_msgs = self.check_actions()
        check_passed, check_msg = self.check_monitoring()
        if not check_passed:
            error_msgs.append(check_msg)
        return error_msgs

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.config_dict == other.config_dict
        return False


def format_volumes(paasta_volume_list):
    return [
        {
            'container_path': v['containerPath'],
            'host_path': v['hostPath'],
            'mode': v['mode'],
        } for v in paasta_volume_list
    ]


def format_master_config(master_config, default_volumes, dockercfg_location):
    mesos_options = master_config.get('mesos_options', {})
    mesos_options.update({
        'default_volumes': format_volumes(default_volumes),
        'dockercfg_location': dockercfg_location,
    })
    master_config['mesos_options'] = mesos_options
    return master_config


def format_tron_action_dict(action_config):
    """Generate a dict of tronfig for an action, from the TronActionConfig.

    :param job_config: TronActionConfig
    """
    executor = action_config.get_executor()
    result = {
        'command': action_config.get_cmd(),
        'executor': executor,
        'requires': action_config.get_requires(),
        'node': action_config.get_node(),
        'retries': action_config.get_retries(),
        'retries_delay': action_config.get_retries_delay(),
        'expected_runtime': action_config.get_expected_runtime(),
        'trigger_downstreams': action_config.get_trigger_downstreams(),
        'triggered_by': action_config.get_triggered_by(),
        'on_upstream_rerun': action_config.get_on_upstream_rerun(),
        'trigger_timeout': action_config.get_trigger_timeout(),
    }
    if executor == 'mesos':
        result['cpus'] = action_config.get_cpus()
        result['mem'] = action_config.get_mem()
        result['env'] = action_config.get_env()
        result['extra_volumes'] = format_volumes(action_config.get_extra_volumes())
        result['docker_parameters'] = [
            {
                'key': param['key'],
                'value': param['value'],
            } for param in action_config.format_docker_parameters()
        ]
        constraint_labels = ['attribute', 'operator', 'value']
        result['constraints'] = [
            dict(zip(constraint_labels, constraint))
            for constraint in action_config.get_calculated_constraints()
        ]

        # If deployments were not loaded
        if not action_config.get_docker_image():
            result['docker_image'] = ''
        else:
            result['docker_image'] = action_config.get_docker_url()

    # Only pass non-None values, so Tron will use defaults for others
    return {key: val for key, val in result.items() if val is not None}


def format_tron_job_dict(job_config):
    """Generate a dict of tronfig for a job, from the TronJobConfig.

    :param job_config: TronJobConfig
    """
    action_dict = {
        action_config.get_action_name(): format_tron_action_dict(action_config)
        for action_config in job_config.get_actions()
    }

    result = {
        'node': job_config.get_node(),
        'schedule': job_config.get_schedule(),
        'actions': action_dict,
        'monitoring': job_config.get_monitoring(),
        'queueing': job_config.get_queueing(),
        'run_limit': job_config.get_run_limit(),
        'all_nodes': job_config.get_all_nodes(),
        'enabled': job_config.get_enabled(),
        'allow_overlap': job_config.get_allow_overlap(),
        'max_runtime': job_config.get_max_runtime(),
        'time_zone': job_config.get_time_zone(),
        'expected_runtime': job_config.get_expected_runtime(),
    }
    cleanup_config = job_config.get_cleanup_action()
    if cleanup_config:
        cleanup_action = format_tron_action_dict(cleanup_config)
        result['cleanup_action'] = cleanup_action

    # Only pass non-None values, so Tron will use defaults for others
    return {key: val for key, val in result.items() if val is not None}


def load_tron_instance_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> TronActionConfig:
    jobs, _ = load_tron_service_config(
        service=service,
        cluster=cluster,
        load_deployments=load_deployments,
        soa_dir=soa_dir,
    )
    requested_job, requested_action = instance.split('.')
    for job in jobs:
        if job.get_name() == requested_job:
            for action in job.get_actions():
                if action.get_action_name() == requested_action:
                    return action
    raise NoConfigurationForServiceError(f"No tron configuration found for {service} {instance}")


def load_tron_yaml(service: str, cluster: str, soa_dir: str) -> Dict[str, Any]:
    tronfig_folder = get_tronfig_folder(soa_dir=soa_dir, cluster=cluster)
    config = service_configuration_lib.read_extra_service_information(
        service_name=service,
        extra_info=f'tron-{cluster}',
        soa_dir=soa_dir,
    )
    if not config:
        config = service_configuration_lib._read_yaml_file(os.path.join(tronfig_folder, f"{service}.yaml"))
    if not config:
        raise NoConfigurationForServiceError('No Tron configuration found for service %s' % service)
    return config


def load_tron_service_config(service, cluster, load_deployments=True, soa_dir=DEFAULT_SOA_DIR):
    """Load all configured jobs for a service, and any additional config values."""
    config = load_tron_yaml(service=service, cluster=cluster, soa_dir=soa_dir)
    config = {key: value for key, value in config.items() if not key.startswith('_')}  # filter templates
    extra_config = {key: value for key, value in config.items() if key != 'jobs'}
    jobs = config.get('jobs') or []
    if isinstance(jobs, list):
        job_configs = [
            TronJobConfig(
                name=job.get('name'),
                service=service,
                cluster=cluster,
                config_dict=job,
                load_deployments=load_deployments,
                soa_dir=soa_dir,
            ) for job in jobs
        ]
    else:
        job_configs = [
            TronJobConfig(
                name=name,
                service=service,
                cluster=cluster,
                config_dict=job,
                load_deployments=load_deployments,
                soa_dir=soa_dir,
            ) for name, job in jobs.items()
        ]
    return job_configs, extra_config


def create_complete_config(service, cluster, soa_dir=DEFAULT_SOA_DIR):
    """Generate a namespace configuration file for Tron, for a service."""
    system_paasta_config = load_system_paasta_config()

    job_configs, other_config = load_tron_service_config(
        service=service,
        cluster=cluster,
        load_deployments=True,
        soa_dir=soa_dir,
    )

    if service == MASTER_NAMESPACE:
        other_config = format_master_config(
            other_config,
            system_paasta_config.get_volumes(),
            system_paasta_config.get_dockercfg_location(),
        )

    other_config['jobs'] = {
        job_config.get_name(): format_tron_job_dict(job_config)
        for job_config in job_configs
    }

    return yaml.dump(
        other_config,
        Dumper=Dumper,
        default_flow_style=False,
    )


def validate_complete_config(service: str, cluster: str, soa_dir: str = DEFAULT_SOA_DIR) -> List[str]:
    job_configs, other_config = load_tron_service_config(
        service=service,
        cluster=cluster,
        load_deployments=False,
        soa_dir=soa_dir,
    )

    if service != MASTER_NAMESPACE and other_config:
        other_keys = list(other_config.keys())
        return [
            f'Non-{MASTER_NAMESPACE} namespace cannot have other config values, found {other_keys}',
        ]

    # PaaSTA-specific validation
    for job_config in job_configs:
        check_msgs = job_config.validate()
        if check_msgs:
            return check_msgs

    # Use Tronfig on generated config from PaaSTA to validate the rest
    other_config['jobs'] = {
        job_config.get_name(): format_tron_job_dict(job_config)
        for job_config in job_configs
    }

    complete_config = yaml.dump(other_config, Dumper=Dumper)

    master_config_path = os.path.join(os.path.abspath(soa_dir), 'tron', cluster, MASTER_NAMESPACE + '.yaml')
    proc = subprocess.run(
        ['tronfig', '-', '-V', '-n', service, '-m', master_config_path],
        input=complete_config,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding='utf-8',
    )

    if proc.returncode != 0:
        process_errors = proc.stderr.strip()
        if process_errors:  # Error running tronfig
            paasta_print(proc.stderr)
        return [proc.stdout.strip()]

    return []


def _get_tron_namespaces_from_service_dir(cluster, soa_dir):
    tron_config_file = f'tron-{cluster}.yaml'
    config_dirs = [_dir[0] for _dir in os.walk(os.path.abspath(soa_dir)) if tron_config_file in _dir[2]]
    namespaces = [os.path.split(config_dir)[1] for config_dir in config_dirs]
    return namespaces


def _get_tron_namespaces_from_tron_dir(cluster, soa_dir):
    config_dir = os.path.join(
        os.path.abspath(soa_dir),
        'tron',
        cluster,
    )
    namespaces = [
        os.path.splitext(filename)[0] for filename in os.listdir(config_dir)
    ]
    return namespaces


class ConflictingNamespacesError(RuntimeError):
    pass


def get_tron_namespaces_for_cluster(cluster=None, soa_dir=DEFAULT_SOA_DIR):
    """Get all the namespaces that are configured in a particular Tron cluster."""
    if not cluster:
        cluster = load_tron_config().get_cluster_name()

    namespaces1 = set(_get_tron_namespaces_from_service_dir(cluster, soa_dir))
    namespaces2 = set(_get_tron_namespaces_from_tron_dir(cluster, soa_dir))

    if namespaces1.intersection(namespaces2):
        raise ConflictingNamespacesError(
            "namespaces found in both service/*/tron and service/tron/*: {}".
            format(namespaces1.intersection(namespaces2)),
        )

    namespaces = list(namespaces1.union(namespaces2))
    return namespaces


def list_tron_clusters(service: str, soa_dir: str = DEFAULT_SOA_DIR) -> List[str]:
    """Returns the Tron clusters a service is configured to deploy to."""
    search_re = r'/tron-([0-9a-z-_]*)\.yaml$'
    service_dir = os.path.join(soa_dir, service)
    clusters = []
    for filename in glob.glob(f'{service_dir}/*.yaml'):
        cluster_re_match = re.search(search_re, filename)
        if cluster_re_match is not None:
            clusters.append(cluster_re_match.group(1))
    return clusters



#!/usr/bin/env python
""" Meant to be used by mesos-slave instead of the /usr/bin/docker executable
directly This will parse the CLI arguments intended for docker, extract
environment variable settings related to the actual node hostname and mesos
task ID, and use those as an additional --hostname argument when calling the
underlying docker command.

If the environment variables are unspecified, or if --hostname is already
specified, this does not change any arguments and just directly calls docker
as-is.

Additionally this wrapper will look for the environment variable
PIN_TO_NUMA_NODE which contains the physical CPU and memory to restrict the
container to. If the system is NUMA enabled, docker will be called with the
arguments cpuset-cpus and cpuset-mems.
"""
import logging
import os
import re
import socket
import sys

from paasta_tools.firewall import DEFAULT_SYNAPSE_SERVICE_DIR
from paasta_tools.firewall import firewall_flock
from paasta_tools.firewall import prepare_new_container
from paasta_tools.mac_address import reserve_unique_mac_address
from paasta_tools.utils import DEFAULT_SOA_DIR


LOCK_DIRECTORY = '/var/lib/paasta/mac-address'
ENV_MATCH_RE = re.compile(r'^(-\w*e\w*|--env(?P<file>-file)?)(=(?P<arg>\S.*))?$')
MAX_HOSTNAME_LENGTH = 63


def parse_env_args(args):
    result = dict(os.environ.items())
    in_env = False
    in_file = False
    for arg in args:
        if not in_env:
            match = ENV_MATCH_RE.match(arg)
            if not match:
                continue
            arg = match.group('arg') or ''
            in_file = bool(match.group('file'))
            if not arg:
                in_env = True
                continue

        in_env = False

        if in_file:
            result.update(read_env_file(arg))
            in_file = False
            continue

        try:
            k, v = arg.split('=', 1)
        except ValueError:
            continue

        result[k] = v

    return result


def read_env_file(filename):
    # Parse a file where each line is KEY=VALUE
    # return contents in dictionary form
    result = {}
    with open(filename) as f:
        for line in f:
            try:
                k, v = line.split('=', 1)
            except ValueError:
                continue
            result[k] = v.strip()
    return result


def can_add_hostname(args):
    # return False if --hostname is already specified or if --network=host
    if is_network_host(args):
        return False

    for index, arg in enumerate(args):

        # Check for --hostname and variants
        if arg == '-h':
            return False
        if arg.startswith('--hostname'):
            return False
        if len(arg) > 1 and arg[0] == '-' and arg[1] != '-':
            # several short args
            arg = arg.partition('=')[0]
            if 'h' in arg:
                return False

    return True


def is_network_host(args):
    for index, arg in enumerate(args):
        # Check for --network=host and variants
        if arg in ('--net=host', '--network=host'):
            return True
        try:
            if arg in ('--net', '--network') and args[index + 1] == 'host':
                return True
        except IndexError:
            pass

    return False


def is_run(args):
    try:
        list(args).index('run')
        return True
    except ValueError:
        return False


def can_add_mac_address(args):
    # return False if --mac-address is already specified or if --network=host
    if is_network_host(args) or not is_run(args):
        return False

    for index, arg in enumerate(args):
        # Check for --mac-address
        if arg.startswith('--mac-address'):
            return False

    return True


def generate_hostname(fqdn, mesos_task_id):
    host_hostname = fqdn.partition('.')[0]
    task_id = mesos_task_id.rpartition('.')[2]

    hostname = host_hostname + '-' + task_id

    # hostnames can only contain alphanumerics and dashes and must be no more
    # than 63 characters
    hostname = re.sub('[^a-zA-Z0-9-]+', '-', hostname)[:MAX_HOSTNAME_LENGTH]

    # hostnames can also not end with dashes as per RFC952
    hostname = hostname.rstrip('-')

    return hostname


def add_argument(args, argument):
    # Add an argument immediately after 'run' command if it exists
    args = list(args)
    try:
        run_index = args.index('run')
    except ValueError:
        pass
    else:
        args.insert(run_index + 1, argument)
    return args


def get_cpumap():
    # Return a dict containing the core numbers per physical CPU
    core = 0
    cpumap = {}
    try:
        with open('/proc/cpuinfo', 'r') as f:
            for line in f:
                m = re.match(r'physical\sid.*(\d)', line)
                if m:
                    cpuid = int(m.group(1))
                    if cpuid not in cpumap:
                        cpumap[cpuid] = []
                    cpumap[cpuid].append(core)
                    core += 1
    except IOError:
        logging.warning('Error while trying to read cpuinfo')
        pass
    return cpumap


def get_numa_memsize(nb_nodes):
    # Return memory size in mB per NUMA node assuming memory is split evenly
    # TODO: calculate and return real memory map
    try:
        with open('/proc/meminfo', 'r') as f:
            for line in f:
                m = re.match(r'MemTotal:\s*(\d+)\skB', line)
                if m:
                    return int(m.group(1)) / 1024 / int(nb_nodes)
    except IOError:
        logging.warning('Error while trying to read meminfo')
        pass
    return 0


def arg_collision(new_args, current_args):
    # Returns True if one of the new arguments is already in the
    # current argument list.
    cur_arg_keys = []
    for c in current_args:
        cur_arg_keys.append(c.split('=')[0])
    return bool(set(new_args).intersection(set(cur_arg_keys)))


def is_numa_enabled():
    if os.path.exists('/proc/1/numa_maps'):
        return True
    else:
        logging.warning('The system does not support NUMA')
        return False


def get_cpu_requirements(env_args):
    # Ensure we return a float. If no requirements we return 0.0
    try:
        return float(env_args.get('MARATHON_APP_RESOURCE_CPUS'))
    except (ValueError, TypeError):
        logging.warning('Could not read {} as a float'.format(env_args.get('MARATHON_APP_RESOURCE_CPUS')))
        return 0.0


def get_mem_requirements(env_args):
    # Ensure we return a float. If no requirements we return 0.0
    try:
        return float(env_args.get('MARATHON_APP_RESOURCE_MEM'))
    except (ValueError, TypeError):
        logging.warning('Could not read {} as a float'.format(env_args.get('MARATHON_APP_RESOURCE_MEM')))
        return 0.0


def append_cpuset_args(argv, env_args):
    # Enable log messages to stderr
    logging.basicConfig(stream=sys.stderr, level=logging.INFO)

    try:
        pinned_numa_node = int(env_args.get('PIN_TO_NUMA_NODE'))
    except (ValueError, TypeError):
        logging.error('Could not read PIN_TO_NUMA_NODE value as an int: {}'.format(
            env_args.get('PIN_TO_NUMA_NODE'),
        ))
        return argv

    cpumap = get_cpumap()

    if len(cpumap) < 1:
        logging.error('Less than 2 physical CPU detected')
        return argv
    if pinned_numa_node not in cpumap:
        logging.error('Specified NUMA node: {} does not exist on this system'.format(
            pinned_numa_node,
        ))
        return argv
    if arg_collision(['--cpuset-cpus', '--cpuset-mems'], argv):
        logging.error('--cpuset options are already set. Not overriding')
        return argv
    if not is_numa_enabled():
        logging.error('Could not detect NUMA on the system')
        return argv
    if len(cpumap[pinned_numa_node]) < get_cpu_requirements(env_args):
        logging.error('The NUMA node has less cores than requested')
        return argv
    if get_numa_memsize(len(cpumap)) <= get_mem_requirements(env_args):
        logging.error('Requested memory:{} MB does not fit in one NUMA node: {} MB'.format(
            get_mem_requirements(env_args), get_numa_memsize(len(cpumap)),
        ))
        return argv

    logging.info(f'Binding container to NUMA node {pinned_numa_node}')
    argv = add_argument(
        argv, ('--cpuset-cpus=' + ','.join(
            str(c) for c in cpumap[pinned_numa_node]
        )),
    )
    argv = add_argument(argv, ('--cpuset-mems=' + str(pinned_numa_node)))
    return argv


def add_firewall(argv, service, instance):
    output = ''
    try:
        mac_address, lockfile = reserve_unique_mac_address(LOCK_DIRECTORY)
    except Exception as e:
        output = f'Unable to add mac address: {e}'
    else:
        argv = add_argument(argv, f'--mac-address={mac_address}')
        try:

            with firewall_flock():
                prepare_new_container(
                    DEFAULT_SOA_DIR,
                    DEFAULT_SYNAPSE_SERVICE_DIR,
                    service,
                    instance,
                    mac_address,
                )
        except Exception as e:
            output = f'Unable to add firewall rules: {e}'

    if output:
        print(output, file=sys.stderr)

    return argv


def main(argv=None):
    argv = argv if argv is not None else sys.argv

    env_args = parse_env_args(argv)

    if env_args.get('PIN_TO_NUMA_NODE'):
        argv = append_cpuset_args(argv, env_args)

    # Marathon sets MESOS_TASK_ID whereas Chronos sets mesos_task_id
    mesos_task_id = env_args.get('MESOS_TASK_ID') or env_args.get('mesos_task_id')

    if mesos_task_id and can_add_hostname(argv):
        hostname = generate_hostname(socket.getfqdn(), mesos_task_id)
        argv = add_argument(argv, f'--hostname={hostname}')

    paasta_firewall = env_args.get('PAASTA_FIREWALL')
    service = env_args.get('PAASTA_SERVICE')
    instance = env_args.get('PAASTA_INSTANCE')
    if paasta_firewall and service and instance and can_add_mac_address(argv):
        try:
            argv = add_firewall(argv, service, instance)
        except Exception as e:
            print(f'Unhandled exception in add_firewall: {e}', file=sys.stderr)

    os.execlp('docker', 'docker', *argv[1:])



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./cleanup_tron_namespaces.py [options]

Delete namespaces that aren't configured in SOA configs for a Tron cluster.

Gets the list of namespaces from Tron, then compares to the namespaces
defined in SOA configs.

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- --dry-run: Print namespaces to be deleted instead of deleting them
"""
import argparse
import sys

from paasta_tools import tron_tools
from paasta_tools.utils import paasta_print


def parse_args():
    parser = argparse.ArgumentParser(description='Cleans up stale Tron namespaces.')
    parser.add_argument(
        '-d', '--soa-dir', dest='soa_dir', metavar='SOA_DIR',
        default=tron_tools.DEFAULT_SOA_DIR,
        help='Use a different soa config directory',
    )
    parser.add_argument(
        '--dry-run', dest='dry_run', action='store_true',
        help='Print namespaces to be deleted, instead of deleting them',
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    client = tron_tools.get_tron_client()
    namespaces = client.list_namespaces()
    expected_namespaces = tron_tools.get_tron_namespaces_for_cluster(soa_dir=args.soa_dir)
    to_delete = set(namespaces) - set(expected_namespaces)

    if not to_delete:
        paasta_print('No Tron namespaces to remove')
        sys.exit(0)

    if args.dry_run:
        paasta_print('Dry run, would have removed namespaces:\n  ' + '\n  '.join(to_delete))
        sys.exit(0)

    successes = []
    errors = []
    for namespace in to_delete:
        try:
            client.update_namespace(namespace, '')
            successes.append(namespace)
        except Exception as e:
            errors.append((namespace, e))

    if successes:
        paasta_print('Successfully removed namespaces:\n', '\n  '.join(successes))

    if errors:
        paasta_print(
            'Failed to remove namespaces:\n  ' + '\n  '.join(
                ['{namespace}: {error}'.format(namespace=namespace, error=str(error)) for namespace, error in errors],
            ),
        )
        sys.exit(1)


if __name__ == '__main__':
    main()



import os
from typing import Any
from typing import Dict
from typing import List

from service_configuration_lib import read_service_configuration


class BaseSecretProvider:

    def __init__(
        self,
        soa_dir: str,
        service_name: str,
        cluster_names: List[str],
        **kwargs: Any,
    ) -> None:
        self.soa_dir = soa_dir
        self.service_name = service_name
        self.secret_dir = os.path.join(self.soa_dir, self.service_name, "secrets")
        self.cluster_names = cluster_names
        service_config = read_service_configuration(self.service_name, self.soa_dir)
        self.encryption_key = service_config.get('encryption_key', 'paasta')

    def decrypt_environment(self, environment: Dict[str, str], **kwargs: Any) -> Dict[str, str]:
        raise NotImplementedError

    def write_secret(self, action: str, secret_name: str, plaintext: bytes) -> None:
        raise NotImplementedError

    def decrypt_secret(self, secret_name: str) -> str:
        raise NotImplementedError


class SecretProvider(BaseSecretProvider):
    pass



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_kubernetes_job.py <service.instance> [options]

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -v, --verbose: Verbose output
"""
import argparse
import logging
import sys
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from kubernetes.client import V1beta1PodDisruptionBudget
from kubernetes.client import V1DeleteOptions
from kubernetes.client import V1Deployment
from kubernetes.client import V1StatefulSet
from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import create_deployment
from paasta_tools.kubernetes_tools import create_pod_disruption_budget
from paasta_tools.kubernetes_tools import create_stateful_set
from paasta_tools.kubernetes_tools import ensure_paasta_namespace
from paasta_tools.kubernetes_tools import InvalidKubernetesConfig
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import KubeDeployment
from paasta_tools.kubernetes_tools import list_all_deployments
from paasta_tools.kubernetes_tools import load_kubernetes_service_config_no_cache
from paasta_tools.kubernetes_tools import max_unavailable
from paasta_tools.kubernetes_tools import pod_disruption_budget_for_service_instance
from paasta_tools.kubernetes_tools import update_deployment
from paasta_tools.kubernetes_tools import update_stateful_set
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import NoDeploymentsAvailable
from paasta_tools.utils import SPACER

log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Creates marathon jobs.')
    parser.add_argument(
        'service_instance_list', nargs='+',
        help="The list of marathon service instances to create or update",
        metavar="SERVICE%sINSTANCE" % SPACER,
    )
    parser.add_argument(
        '-d', '--soa-dir', dest="soa_dir", metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        dest="verbose", default=False,
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    soa_dir = args.soa_dir
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    # system_paasta_config = load_system_paasta_config()
    kube_client = KubeClient()

    ensure_paasta_namespace(kube_client)
    setup_kube_succeeded = setup_kube_deployments(
        kube_client=kube_client,
        service_instances=args.service_instance_list,
        soa_dir=soa_dir,
    )
    sys.exit(0 if setup_kube_succeeded else 1)


def setup_kube_deployments(
    kube_client: KubeClient,
    service_instances: Sequence[str],
    soa_dir: str = DEFAULT_SOA_DIR,
) -> bool:
    succeeded = True
    if service_instances:
        deployments = list_all_deployments(kube_client)
    for service_instance in service_instances:
        try:
            service, instance, _, __ = decompose_job_id(service_instance)
        except InvalidJobNameError:
            log.error("Invalid service instance specified. Format is service%sinstance." % SPACER)
            succeeded = False
        else:
            if reconcile_kubernetes_deployment(
                kube_client=kube_client,
                service=service,
                instance=instance,
                kube_deployments=deployments,
                soa_dir=soa_dir,
            )[0]:
                succeeded = False
    return succeeded


def reconcile_kubernetes_deployment(
    kube_client: KubeClient,
    service: str,
    instance: str,
    kube_deployments: Sequence[KubeDeployment],
    soa_dir: str,
) -> Tuple[int, Optional[int]]:
    try:
        service_instance_config = load_kubernetes_service_config_no_cache(
            service,
            instance,
            load_system_paasta_config().get_cluster(),
            soa_dir=soa_dir,
        )
    except NoDeploymentsAvailable:
        log.debug("No deployments found for %s.%s in cluster %s. Skipping." %
                  (service, instance, load_system_paasta_config().get_cluster()))
        return 0, None
    except NoConfigurationForServiceError:
        error_msg = "Could not read kubernetes configuration file for %s.%s in cluster %s" % \
                    (service, instance, load_system_paasta_config().get_cluster())
        log.error(error_msg)
        return 1, None

    try:
        formatted_application = service_instance_config.format_kubernetes_app()
    except InvalidKubernetesConfig as e:
        log.error(str(e))
        return (1, None)

    desired_deployment = KubeDeployment(
        service=service,
        instance=instance,
        git_sha=formatted_application.metadata.labels["git_sha"],
        config_sha=formatted_application.metadata.labels["config_sha"],
        replicas=formatted_application.spec.replicas,
    )

    if not (service, instance) in [(kd.service, kd.instance) for kd in kube_deployments]:
        log.debug(f"{desired_deployment} does not exist so creating")
        create_kubernetes_application(
            kube_client=kube_client,
            application=formatted_application,
        )
    elif desired_deployment not in kube_deployments:
        log.debug(f"{desired_deployment} exists but config_sha or git_sha doesn't match or number of instances changed")
        update_kubernetes_application(
            kube_client=kube_client,
            application=formatted_application,
        )
    else:
        log.debug(f"{desired_deployment} is up to date, no action taken")

    ensure_pod_disruption_budget(
        kube_client=kube_client,
        service=service,
        instance=instance,
        min_instances=service_instance_config.get_desired_instances() - max_unavailable(
            instance_count=service_instance_config.get_desired_instances(),
            bounce_margin_factor=service_instance_config.get_bounce_margin_factor(),
        ),
    )
    return 0, None


def ensure_pod_disruption_budget(
        kube_client: KubeClient,
        service: str,
        instance: str,
        min_instances: int,
) -> V1beta1PodDisruptionBudget:
    pdr = pod_disruption_budget_for_service_instance(
        service=service,
        instance=instance,
        min_instances=min_instances,
    )
    try:
        existing_pdr = kube_client.policy.read_namespaced_pod_disruption_budget(
            name=pdr.metadata.name,
            namespace=pdr.metadata.namespace,
        )
    except ApiException as e:
        if e.status == 404:
            existing_pdr = None
        else:
            raise

    if existing_pdr:
        if existing_pdr.spec.min_available != pdr.spec.min_available:
            # poddisruptionbudget objects are not mutable like most things in the kubernetes api,
            # so we have to do a delete/replace.
            # unfortunately we can't really do this transactionally, but I guess we'll just hope for the best?
            logging.debug(f'existing poddisruptionbudget {pdr.metadata.name} is out of date; deleting')
            kube_client.policy.delete_namespaced_pod_disruption_budget(
                name=pdr.metadata.name,
                namespace=pdr.metadata.namespace,
                body=V1DeleteOptions(),
            )
            logging.debug(f'creating poddisruptionbudget {pdr.metadata.name}')
            return create_pod_disruption_budget(
                kube_client=kube_client,
                pod_disruption_budget=pdr,
            )
        else:
            logging.debug(f'poddisruptionbudget {pdr.metadata.name} up to date')
    else:
        logging.debug(f'creating poddisruptionbudget {pdr.metadata.name}')
        return create_pod_disruption_budget(
            kube_client=kube_client,
            pod_disruption_budget=pdr,
        )


def create_kubernetes_application(kube_client: KubeClient, application: Union[V1Deployment, V1StatefulSet]) -> None:
    if isinstance(application, V1Deployment):
        create_deployment(
            kube_client=kube_client,
            formatted_deployment=application,
        )
    elif isinstance(application, V1StatefulSet):
        create_stateful_set(
            kube_client=kube_client,
            formatted_stateful_set=application,
        )
    else:
        raise Exception("Unknown kubernetes object to create")


def update_kubernetes_application(kube_client: KubeClient, application: Union[V1Deployment, V1StatefulSet]) -> None:
    if isinstance(application, V1Deployment):
        update_deployment(
            kube_client=kube_client,
            formatted_deployment=application,
        )
    elif isinstance(application, V1StatefulSet):
        update_stateful_set(
            kube_client=kube_client,
            formatted_stateful_set=application,
        )
    else:
        raise Exception("Unknown kubernetes object to update")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import re
from shlex import quote

from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_clusters
from paasta_tools.cli.utils import list_instances
from paasta_tools.cli.utils import run_on_master
from paasta_tools.utils import list_services
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import PaastaNotConfiguredError
from paasta_tools.utils import SystemPaastaConfig


ARG_DEFAULTS = dict(
    common=dict(
        service=None,
        instance=None,
        cluster=None,  # load from system paasta config later
        verbose=False,
    ),
    start=dict(
        cmd=None,
        detach=False,
        staging_timeout=240.0,
        instances=1,
        docker_image=None,
        dry_run=False,
        constraint=[],
        notification_email=None,
        retries=0,
    ),
    stop=dict(run_id=None, framework_id=None),
    list=dict(),
)


def get_system_paasta_config():
    try:
        return load_system_paasta_config()
    except PaastaNotConfiguredError:
        paasta_print(
            PaastaColors.yellow(
                "Warning: Couldn't load config files from '/etc/paasta'. This "
                "indicates PaaSTA is not configured locally on this host, and "
                "remote-run may not behave the same way it would behave on a "
                "server configured for PaaSTA.",
            ),
            sep='\n',
        )
        return SystemPaastaConfig(
            {"volumes": []},
            '/etc/paasta',
        )


def add_common_args_to_parser(parser):
    parser.add_argument(
        '-s', '--service',
        help='The name of the service you wish to inspect. Required.',
        required=True,
    ).completer = lazy_choices_completer(list_services)
    parser.add_argument(
        '-i', '--instance',
        help=(
            "Simulate a docker run for a particular instance of the "
            "service, like 'main' or 'canary'. Required."
        ),
        required=True,
    ).completer = lazy_choices_completer(list_instances)
    parser.add_argument(
        '-c', '--cluster',
        help=(
            'The name of the cluster you wish to run your task on. '
            'If omitted, uses the default cluster defined in the paasta '
            f'remote-run configs.'
        ),
        default=ARG_DEFAULTS['common']['cluster'],
    ).completer = lazy_choices_completer(list_clusters)
    parser.add_argument(
        '-v', '--verbose',
        help='Show more output',
        action='store_true',
        default=ARG_DEFAULTS['common']['verbose'],
    )


def add_start_parser(subparser):
    parser = subparser.add_parser(
        'start',
        help="Start task subcommand",
    )
    add_common_args_to_parser(parser)
    parser.add_argument(
        '-C', '--cmd',
        help=(
            'Run Docker container with particular command, for example: '
            '"bash". By default will use the command or args specified by the '
            'soa-configs or what was specified in the Dockerfile'
        ),
        default=ARG_DEFAULTS['start']['cmd'],
    ),
    parser.add_argument(
        '-D', '--detach',
        help='Launch in background',
        action='store_true',
        default=ARG_DEFAULTS['start']['detach'],
    )
    default_staging_timeout = ARG_DEFAULTS['start']['staging_timeout']
    parser.add_argument(
        '-t', '--staging-timeout',
        help=(
            'A timeout in seconds for the task to be launching before killed. '
            f'Default: {default_staging_timeout}s'
        ),
        default=ARG_DEFAULTS['start']['staging_timeout'],
        type=float,
    )
    parser.add_argument(
        '-j', '--instances',
        help='Number of copies of the task to launch',
        default=ARG_DEFAULTS['start']['instances'],
        type=int,
    )
    parser.add_argument(
        '--docker-image',
        help=(
            'URL of docker image to use. '
            'Defaults to using the deployed docker image.'
        ),
        default=ARG_DEFAULTS['start']['docker_image'],
    )
    parser.add_argument(
        '-R', '--run-id',
        help='ID of task to stop',
        default=ARG_DEFAULTS['stop']['run_id'],
    )
    parser.add_argument(
        '-d', '--dry-run',
        help=(
            'Don\'t launch the task. '
            'Instead output task that would have been launched'
        ),
        action='store_true',
        default=ARG_DEFAULTS['start']['dry_run'],
    )
    parser.add_argument(
        '-X', '--constraint',
        help='Constraint option, format: <attr>,OP[,<value>], OP can be one '
        'of the following: EQUALS matches attribute value exactly, LIKE and '
        'UNLIKE match on regular expression, MAX_PER constrains number of '
        'tasks per attribute value, UNIQUE is the same as MAX_PER,1',
        action='append',
        default=ARG_DEFAULTS['start']['constraint'],
    )
    default_email = os.environ.get('EMAIL', None)
    parser.add_argument(
        '-E', '--notification-email',
        help=(
            'Email address to send remote-run notifications to. '
            'A notification will be sent when a task either succeeds or fails. '
            'Defaults to env variable $EMAIL: '
        ) + (default_email if default_email else '(currently not set)'),
        default=default_email,
    )
    default_retries = ARG_DEFAULTS['start']['retries']
    parser.add_argument(
        '-r', '--retries',
        help=(
            'Number of times to retry if task fails at launch or at runtime. '
            f'Default: {default_retries}'
        ),
        type=int,
        default=default_retries,
    )
    return parser


def add_stop_parser(subparser):
    parser = subparser.add_parser(
        'stop',
        help="Stop task subcommand",
    )
    add_common_args_to_parser(parser)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '-R', '--run-id',
        help='ID of task to stop',
        default=ARG_DEFAULTS['stop']['run_id'],
    )
    group.add_argument(
        '-F', '--framework-id',
        help=(
            'ID of framework to stop. Must belong to remote-run of selected '
            'service instance.'
        ),
        type=str,
        default=ARG_DEFAULTS['stop']['framework_id'],
    )
    return parser


def add_list_parser(subparser):
    parser = subparser.add_parser(
        'list',
        help="List tasks subcommand",
    )
    add_common_args_to_parser(parser)
    return parser


def add_subparser(subparsers):
    main_parser = subparsers.add_parser(
        'remote-run',
        help="Schedule Mesos to run adhoc command in context of a service",
        description=(
            "`paasta remote-run` is useful for running adhoc commands in "
            "context of a service's Docker image. The command will be "
            "scheduled on a Mesos cluster and stdout/stderr printed after "
            "execution is finished."
        ),
        epilog=(
            "Note: `paasta remote-run` uses Mesos API that may require "
            "authentication."
        ),
    )
    main_subs = main_parser.add_subparsers(
        dest='action',
        help='Subcommands of remote-run',
    )
    add_start_parser(main_subs)
    add_stop_parser(main_subs)
    add_list_parser(main_subs)
    main_parser.set_defaults(command=paasta_remote_run)


def split_constraints(constraints):
    return [c.split(',', 2) for c in constraints]


def create_remote_run_command(args):
    cmd_parts = ['/usr/bin/paasta_remote_run', args.action]
    arg_vars = vars(args)
    arg_defaults = dict(ARG_DEFAULTS[args.action])  # copy dict
    arg_defaults.update(ARG_DEFAULTS['common'])
    arg_defaults.pop('constraint')  # needs conversion to json

    for k in arg_vars.keys():
        if k not in arg_defaults:  # skip keys we don't know about
            continue
        v = arg_vars[k]
        if v == arg_defaults[k]:  # skip values that have default value
            continue
        k = re.sub(r'_', '-', k)
        if isinstance(v, bool) and v:
            cmd_parts.append(f'--{k}')
        else:
            cmd_parts.extend([f'--{k}', quote(str(v))])

    # constraint, convert to json
    if len(arg_vars['constraint']) > 0:
        constraints = split_constraints(arg_vars['constraint'])
        cmd_parts.extend(['--constraints-json', quote(json.dumps(constraints))])

    return cmd_parts


def paasta_remote_run(args):
    system_paasta_config = get_system_paasta_config()

    if not args.cluster:
        default_cluster = system_paasta_config.get_remote_run_config().get('default_cluster')
        if not default_cluster:
            paasta_print(PaastaColors.red(
                "Error: no cluster specified and no default cluster available",
            ))
            return 1
        args.cluster = default_cluster

    cmd_parts = create_remote_run_command(args)
    graceful_exit = (args.action == 'start' and not args.detach)
    return_code, status = run_on_master(
        cluster=args.cluster,
        system_paasta_config=system_paasta_config,
        cmd_parts=cmd_parts,
        graceful_exit=graceful_exit,
    )

    # Status results are streamed. This print is for possible error messages.
    if status is not None:
        for line in status.rstrip().split('\n'):
            paasta_print('    %s' % line)

    return return_code



#!/usr/bin/env python
# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict
from typing import List
from typing import Tuple

from pymesos import MesosSchedulerDriver

from paasta_tools.frameworks.constraints import ConstraintState
from paasta_tools.frameworks.native_scheduler import LIVE_TASK_STATES
from paasta_tools.frameworks.native_scheduler import NativeScheduler
from paasta_tools.frameworks.native_service_config import TaskInfo
from paasta_tools.frameworks.native_service_config import UnknownNativeServiceError
from paasta_tools.utils import paasta_print


class AdhocScheduler(NativeScheduler):
    def __init__(self, *args, **kwargs):
        self.dry_run = kwargs.pop('dry_run')

        if kwargs.get('service_config_overrides') is None:
            kwargs['service_config_overrides'] = {}
        kwargs['service_config_overrides'].setdefault('instances', 1)
        self.finished_countdown = kwargs['service_config_overrides']['instances']

        super().__init__(*args, **kwargs)

    def need_to_stop(self):
        # Is used to decide whether to stop the driver or try to start more tasks.
        return self.finished_countdown == 0

    def statusUpdate(self, driver: MesosSchedulerDriver, update: Dict):
        super().statusUpdate(driver, update)

        if update['state'] not in LIVE_TASK_STATES:
            self.finished_countdown -= 1

        # Stop if task ran and finished
        if self.need_to_stop():
            driver.stop()

    def tasks_and_state_for_offer(
        self,
        driver: MesosSchedulerDriver,
        offer,
        state: ConstraintState,
    ) -> Tuple[List[TaskInfo], ConstraintState]:
        # In dry run satisfy exit-conditions after we got the offer
        if self.dry_run or self.need_to_stop():
            if self.dry_run:
                tasks, _ = super(). \
                    tasks_and_state_for_offer(driver, offer, state)
                paasta_print("Would have launched: ", tasks)
            driver.stop()
            return [], state

        return super(). \
            tasks_and_state_for_offer(driver, offer, state)

    def kill_tasks_if_necessary(self, *args, **kwargs):
        return

    def validate_config(self):
        if self.service_config.get_cmd() is None:
            raise UnknownNativeServiceError("missing cmd in service config")



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# PYTHON_ARGCOMPLETE_OK
"""A command line tool for viewing information from the PaaSTA stack."""
import argparse
import logging
import sys

import argcomplete
import pkg_resources

from paasta_tools.cli import cmds
from paasta_tools.cli.utils import load_method
from paasta_tools.cli.utils import modules_in_pkg as paasta_commands_dir
from paasta_tools.utils import paasta_print


class PrintsHelpOnErrorArgumentParser(argparse.ArgumentParser):
    """Overriding the error method allows us to print the whole help page,
    otherwise the python arg parser prints a not-so-useful usage message that
    is way too terse"""

    def error(self, message):
        paasta_print("Argument parse error: %s" % message)
        self.print_help()
        sys.exit(1)


def add_subparser(command, subparsers):
    """Given a command name, paasta_cmd, execute the add_subparser method
    implemented in paasta_cmd.py.

    Each paasta client command must implement a method called add_subparser.
    This allows the client to dynamically add subparsers to its subparser, which
    provides the benefits of argcomplete/argparse but gets it done in a modular
    fashion.

    :param command: a simple string - e.g. 'list'
    :param subparsers: an ArgumentParser object"""
    module_name = 'paasta_tools.cli.cmds.%s' % command
    add_subparser_fn = load_method(module_name, 'add_subparser')
    add_subparser_fn(subparsers)


def get_argparser():
    parser = PrintsHelpOnErrorArgumentParser(
        description=(
            "The PaaSTA command line tool. The 'paasta' command is the entry point "
            "to multiple subcommands, see below.\n\n"
            "You can see more help for individual commands by appending them with '--help', "
            "for example, 'paasta status --help' or see the man page with 'man paasta status'."
        ),
        epilog=(
            "The 'paasta' command line tool is designed to be used by humans, and therefore has "
            "command line completion for almost all options and uses pretty formatting when "
            "possible."
        ),
        # Suppressing usage prevents it from being printed twice upon print_help
        usage=argparse.SUPPRESS,
    )

    # http://stackoverflow.com/a/8521644/812183
    parser.add_argument(
        '-V', '--version',
        action='version',
        version='paasta-tools {}'.format(
            pkg_resources.get_distribution('paasta-tools').version,
        ),
    )

    subparsers = parser.add_subparsers(help="[-h, --help] for subcommand help", dest='command')
    subparsers.required = True

    # Adding a separate help subparser allows us to respond to "help" without --help
    help_parser = subparsers.add_parser('help', add_help=False)
    help_parser.set_defaults(command=None)

    for command in sorted(paasta_commands_dir(cmds)):
        add_subparser(command, subparsers)

    return parser


def parse_args(argv):
    """Initialize autocompletion and configure the argument parser.

    :return: an argparse.Namespace object mapping parameter names to the inputs
             from sys.argv
    """
    parser = get_argparser()
    argcomplete.autocomplete(parser)

    return parser.parse_args(argv), parser


def main(argv=None):
    """Perform a paasta call. Read args from sys.argv and pass parsed args onto
    appropriate command in paasta_cli/cmds directory.

    Ensure we kill any child pids before we quit
    """
    logging.basicConfig()
    try:
        args, parser = parse_args(argv)
        if args.command is None:
            parser.print_help()
            return_code = 0
        else:
            return_code = args.command(args)
    except KeyboardInterrupt:
        return_code = 1
    sys.exit(return_code)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from chronos import ChronosAPIError

from paasta_tools.chronos_tools import get_chronos_client
from paasta_tools.chronos_tools import load_chronos_config
from paasta_tools.metrics.metastatus_lib import assert_chronos_scheduled_jobs
from paasta_tools.utils import paasta_print


def check_chronos_jobs():
    config = load_chronos_config()
    if not config:
        paasta_print("UNKNOWN: Failed to load chronos config")
        sys.exit(3)
    client = get_chronos_client(config)

    try:
        result = assert_chronos_scheduled_jobs(client)
    except (ChronosAPIError) as e:
        paasta_print("CRITICAL: Unable to connect to Chronos: %s" % e.message)
        sys.exit(2)

    if result.healthy:
        paasta_print("OK: " + result.message)
        sys.exit(0)
    else:
        paasta_print("CRITICAL: " + result.message)
        sys.exit(2)


if __name__ == '__main__':
    check_chronos_jobs()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging

from paasta_tools.autoscaling.autoscaling_service_lib import autoscale_services
from paasta_tools.marathon_tools import DEFAULT_SOA_DIR


def parse_args():
    parser = argparse.ArgumentParser(description='Autoscales marathon jobs')
    parser.add_argument(
        '-d', '--soa-dir', dest="soa_dir", metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help="Increase logging verboseness",
    )
    parser.add_argument(
        'services', type=str, nargs='*',
        help='name of services to scale (optional defaults to all autoscaling enabled services)',
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)
    autoscale_services(soa_dir=args.soa_dir, services=args.services)


if __name__ == '__main__':
    main()



import time
from datetime import datetime

from tzlocal import get_localzone

from paasta_tools.api import client
from paasta_tools.utils import paasta_print


def get_service_autoscale_pause_time(cluster):
    api = client.get_paasta_api_client(cluster=cluster, http_res=True)
    if not api:
        paasta_print('Could not connect to paasta api. Maybe you misspelled the cluster?')
        return 1
    pause_time, http = api.service_autoscaler.get_service_autoscaler_pause().result()
    if http.status_code == 500:
        paasta_print('Could not connect to zookeeper server')
        return 2

    pause_time = float(pause_time)
    if pause_time < time.time():
        paasta_print('Service autoscaler is not paused')
    else:
        local_tz = get_localzone()
        paused_readable = local_tz.localize(datetime.fromtimestamp(pause_time)).strftime('%F %H:%M:%S %Z')
        paasta_print(f'Service autoscaler is paused until {paused_readable}')

    return 0


def update_service_autoscale_pause_time(cluster, mins):
    api = client.get_paasta_api_client(cluster=cluster, http_res=True)
    if not api:
        paasta_print('Could not connect to paasta api. Maybe you misspelled the cluster?')
        return 1
    body = {'minutes': mins}
    res, http = api.service_autoscaler.update_service_autoscaler_pause(json_body=body).result()
    if http.status_code == 500:
        paasta_print('Could not connect to zookeeper server')
        return 2

    paasta_print(f'Service autoscaler is paused for {mins}')
    return 0


def delete_service_autoscale_pause_time(cluster):
    api = client.get_paasta_api_client(cluster=cluster, http_res=True)
    if not api:
        paasta_print('Could not connect to paasta api. Maybe you misspelled the cluster?')
        return 1
    res, http = api.service_autoscaler.delete_service_autoscaler_pause().result()
    if http.status_code == 500:
        paasta_print('Could not connect to zookeeper server')
        return 2

    paasta_print('Service autoscaler is unpaused')
    return 0



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import time
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Type
from typing import TypeVar

import aiohttp
from mypy_extensions import TypedDict

from paasta_tools.hacheck import get_spool
from paasta_tools.hacheck import post_spool
from paasta_tools.utils import get_user_agent

_drain_methods: Dict[str, Type["DrainMethod"]] = {}


_RegisterDrainMethod_T = TypeVar('_RegisterDrainMethod_T', bound=Type["DrainMethod"])
T = TypeVar('T')


def register_drain_method(name: str) -> Callable[[_RegisterDrainMethod_T], _RegisterDrainMethod_T]:
    """Returns a decorator that registers a DrainMethod subclass at a given name
    so get_drain_method/list_drain_methods can find it."""
    def outer(drain_method: _RegisterDrainMethod_T) -> _RegisterDrainMethod_T:
        _drain_methods[name] = drain_method
        return drain_method
    return outer


def get_drain_method(
    name: str,
    service: str,
    instance: str,
    registrations: List[str],
    drain_method_params: Optional[Dict] = None,
) -> "DrainMethod":
    return _drain_methods[name](service, instance, registrations, **(drain_method_params or {}))


def list_drain_methods() -> List[str]:
    return sorted(_drain_methods.keys())


DrainTask = TypeVar('DrainTask', bound=Any)


class DrainMethod:
    """A drain method is a way of stopping new traffic to tasks without killing them. For example, you might take a task
    out of a load balancer by causing its healthchecks to fail.

    A drain method must have the following methods:
     - drain(task): Begin draining traffic from a task. This should be idempotent.
     - stop_draining(task): Stop draining traffic from a task. This should be idempotent.
     - is_draining(task): Whether a task has already been marked as downed. Note that this state should be stored out of
                          process, because a bounce may take multiple runs of setup_marathon_job to complete.
     - is_safe_to_kill(task): Return True if this task is safe to kill, False otherwise.

    When implementing a drain method, be sure to decorate with @register_drain_method(name).
    """

    def __init__(self, service: str, instance: str, registrations: List[str], **kwargs: Dict) -> None:
        self.service = service
        self.instance = instance
        self.registrations = registrations

    async def drain(self, task: DrainTask) -> None:
        """Make a task stop receiving new traffic."""
        raise NotImplementedError()

    async def stop_draining(self, task: DrainTask) -> None:
        """Make a task that has previously been downed start receiving traffic again."""
        raise NotImplementedError()

    async def is_draining(self, task: DrainTask) -> bool:
        """Return whether a task is being drained."""
        raise NotImplementedError()

    async def is_safe_to_kill(self, task: DrainTask) -> bool:
        """Return True if a task is drained and ready to be killed, or False if we should wait."""
        raise NotImplementedError()


@register_drain_method('noop')
class NoopDrainMethod(DrainMethod):
    """This drain policy does nothing and assumes every task is safe to kill."""

    async def drain(self, task: DrainTask) -> None:
        pass

    async def stop_draining(self, task: DrainTask) -> None:
        pass

    async def is_draining(self, task: DrainTask) -> bool:
        return False

    async def is_safe_to_kill(self, task: DrainTask) -> bool:
        return True


@register_drain_method('test')
class TestDrainMethod(DrainMethod):
    """This drain policy is meant for integration testing. Do not use."""

    # These are variables on the class for ease of use in testing.
    downed_task_ids: Set[str] = set()
    safe_to_kill_task_ids: Set[str] = set()

    async def drain(self, task: DrainTask) -> None:
        if task.id not in self.safe_to_kill_task_ids:
            self.downed_task_ids.add(task.id)

    async def stop_draining(self, task: DrainTask) -> None:
        self.downed_task_ids -= {task.id}
        self.safe_to_kill_task_ids -= {task.id}

    async def is_draining(self, task: DrainTask) -> bool:
        return task.id in (self.downed_task_ids | self.safe_to_kill_task_ids)

    async def is_safe_to_kill(self, task: DrainTask) -> bool:
        return task.id in self.safe_to_kill_task_ids

    @classmethod
    def mark_arbitrary_task_as_safe_to_kill(cls) -> None:
        cls.safe_to_kill_task_ids.add(cls.downed_task_ids.pop())


@register_drain_method('crashy_drain')
class CrashyDrainDrainMethod(NoopDrainMethod):
    async def drain(self, task: DrainTask) -> None:
        raise Exception("Intentionally crashing for testing purposes")


@register_drain_method('crashy_is_safe_to_kill')
class CrashySafeToKillDrainMethod(NoopDrainMethod):
    async def is_safe_to_kill(self, task: DrainTask) -> bool:
        raise Exception("Intentionally crashing for testing purposes")


@register_drain_method('hacheck')
class HacheckDrainMethod(DrainMethod):
    """This drain policy issues a POST to hacheck's /spool/{service}/{port}/status endpoint to cause healthchecks to
    fail. It considers tasks safe to kill if they've been down in hacheck for more than a specified delay."""

    def __init__(
        self,
        service: str,
        instance: str,
        registrations: List[str],
        delay: float = 240,
        hacheck_port: int = 6666,
        expiration: float = 0,
        **kwargs: Dict,
    ) -> None:
        super().__init__(service, instance, registrations)
        self.delay = float(delay)
        self.hacheck_port = hacheck_port
        self.expiration = float(expiration) or float(delay) * 10

    def spool_urls(self, task: DrainTask) -> List[str]:
        return [
            'http://%(task_host)s:%(hacheck_port)d/spool/%(registration)s/%(task_port)d/status' % {
                'task_host': task.host,
                'task_port': task.ports[0],
                'hacheck_port': self.hacheck_port,
                'registration': registration,
            } for registration in self.registrations
        ]

    async def for_each_registration(
        self,
        task: DrainTask,
        func: Callable[..., Awaitable[T]]
    ) -> 'asyncio.Future[T]':
        if task.ports == []:
            return None
        futures = [
            func(url)
            for url in self.spool_urls(task)
        ]
        return await asyncio.gather(*futures)

    async def drain(self, task: DrainTask) -> None:
        await self.for_each_registration(
            task,
            self.down,
        )

    async def up(self, url: str) -> None:
        await post_spool(
            url=url,
            status='up',
            data={
                'status': 'up',
            },
        )

    async def down(self, url: str) -> None:
        await post_spool(
            url=url,
            status='down',
            data={
                'status': 'down',
                'expiration': str(time.time() + self.expiration),
                'reason': 'Drained by Paasta',
            },
        )

    async def stop_draining(self, task: DrainTask) -> None:
        await self.for_each_registration(
            task,
            self.up,
        )

    async def is_draining(self, task: DrainTask) -> bool:
        results = await self.for_each_registration(
            task,
            get_spool,
        )
        return not all([
            res is None or res["state"] == "up"
            for res in results
        ])

    async def is_safe_to_kill(self, task: DrainTask) -> bool:
        results = await self.for_each_registration(
            task,
            lambda url: get_spool(url),
        )
        if all([res is None or res["state"] == "up" for res in results]):
            return False
        else:
            return all([
                res.get("since", 0) < (time.time() - self.delay)
                for res in results
            ])


class StatusCodeNotAcceptableError(Exception):
    pass


UrlSpec = TypedDict(
    'UrlSpec',
    {
        'url_format': str,
        'method': str,
        'success_codes': str,
    },
    total=False,
)


@register_drain_method('http')
class HTTPDrainMethod(DrainMethod):
    """This drain policy issues arbitrary HTTP calls to arbitrary URLs specified by the parameters. The URLs are
    specified as format strings, and will have variables such as {host}, {port}, etc. filled in."""

    def __init__(
        self,
        service: str,
        instance: str,
        registrations: List[str],
        drain: UrlSpec,
        stop_draining: UrlSpec,
        is_draining: UrlSpec,
        is_safe_to_kill: UrlSpec,
    ) -> None:
        super().__init__(service, instance, registrations)
        self.drain_url_spec = drain
        self.stop_draining_url_spec = stop_draining
        self.is_draining_url_spec = is_draining
        self.is_safe_to_kill_url_spec = is_safe_to_kill

    def get_format_params(self, task: DrainTask) -> List[Dict[str, Any]]:
        return [
            {
                'host': task.host,
                'port': task.ports[0],
                'service': self.service,
                'instance': self.instance,
                'nerve_ns': nerve_ns,
            } for nerve_ns in self.registrations
        ]

    def format_url(self, url_format: str, format_params: Dict[str, Any]) -> str:
        return url_format.format(**format_params)

    def parse_success_codes(self, success_codes_str: str) -> Set[int]:
        """Expand a string like 200-399,407-409,500 to a set containing all the integers in between."""
        acceptable_response_codes: Set[int] = set()
        for series_str in str(success_codes_str).split(','):
            if '-' in series_str:
                start, end = series_str.split('-')
                acceptable_response_codes.update(range(int(start), int(end) + 1))
            else:
                acceptable_response_codes.add(int(series_str))
        return acceptable_response_codes

    def check_response_code(self, status_code: int, success_codes_str: str) -> None:
        acceptable_response_codes = self.parse_success_codes(success_codes_str)
        if status_code not in acceptable_response_codes:
            raise StatusCodeNotAcceptableError("Status code %d not in %s", status_code, success_codes_str)

    async def issue_request(self, url_spec: UrlSpec, task: DrainTask) -> None:
        """Issue a request to the URL specified by url_spec regarding the task given."""
        format_params = self.get_format_params(task)
        urls = [self.format_url(url_spec['url_format'], param) for param in format_params]
        method = url_spec.get('method', 'GET').upper()

        async with aiohttp.ClientSession() as session:
            reqs = [
                session.request(
                    method=method,
                    url=url,
                    headers={'User-Agent': get_user_agent()},
                    timeout=15,
                ) for url in urls
            ]
            res = await asyncio.gather(*reqs)
            for response in res:
                self.check_response_code(response.status, url_spec['success_codes'])

    async def drain(self, task: DrainTask) -> None:
        return await self.issue_request(self.drain_url_spec, task)

    async def stop_draining(self, task: DrainTask) -> None:
        return await self.issue_request(self.stop_draining_url_spec, task)

    async def is_draining(self, task: DrainTask) -> bool:
        try:
            await self.issue_request(self.is_draining_url_spec, task)
        except StatusCodeNotAcceptableError:
            return False
        else:
            return True

    async def is_safe_to_kill(self, task: DrainTask) -> bool:
        try:
            await self.issue_request(self.is_safe_to_kill_url_spec, task)
        except StatusCodeNotAcceptableError:
            return False
        else:
            return True
